# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================
def get_ida_taint_overlay(entrypoint, trace, taint_graph, file):
    import json

    # Find all addresses that are not sources of another address.
    # These should be highlighted as the beginning of the taint
    # Or end of reverse taint. However you want to look at it.
    addr_seen = {}
    for idx, sources in taint_graph._path.items():
        addr = trace._trace[idx]
        addr_seen[addr] = False
        for src in sources:
            addr_seen[src[0]] = True

    out_map = {
        "entrypoint": entrypoint,
        "comments": [],
    }
    for addr, src in path.items():
        comment = ", ".join([f"0x{x[0]:x}|{x[1]}" for x in src])
        comment_struct = {
            "address": addr,
            "thread_id": "main",
            "text": f" <- {comment}",
        }
        if src[0][1] == "user_taint":
            comment_struct["color"] = 0xE16563
        if not addr_seen[addr]:
            comment_struct["color"] = 0x81CD4E

        # green

        out_map["comments"].append(comment_struct)

    r = json.dumps(out_map)
    loaded_r = json.loads(r)
    file.write("DISAS\n" + json.dumps(loaded_r, indent=4, sort_keys=True))


# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================
from collections import defaultdict
import html


def render_path_taint_graph(logger, taint_graph, trace, max_nodes=50):
    try:
        from graphviz import Digraph
    except Exception as e:
        return

    logger.info("Creating full taint graphviz")

    graph = Digraph(filename="cfg", format="png")

    included_keys = set()
    to_process = [trace._last_idx_of_address(taint_graph._start_addr)]
    while len(included_keys) < max_nodes and len(to_process) > 0:
        idx = to_process.pop(0)
        if idx in included_keys:
            continue
        graph.node(
            f"{idx:x}",
            taint_graph._addr_label(trace._trace[idx]),
            style="filled",
        )
        parents = taint_graph._path_parents.get(idx, [])
        to_process.extend(parents)
        included_keys.add(idx)

    for key, srcs in taint_graph._path.items():
        for src_key, src in srcs.items():
            if key not in included_keys or src_key not in included_keys:
                continue
            graph.edge(f"{key:x}", f"{src_key:x}", label=src.use_string())

    graph.render()


def render_reduced_path_taint_graph(logger, taint_graph, max_nodes=50):
    try:
        from graphviz import Digraph
    except Exception as e:
        return

    logger.info("Creating reduced taint graphviz")

    graph = Digraph(filename="reduced_cfg", format="png")

    included_keys = set()
    to_process = [taint_graph._start_addr]
    while len(included_keys) < max_nodes and len(to_process) > 0:
        addr = to_process.pop(0)
        if addr in included_keys:
            continue
        graph.node(
            f"{addr:x}", taint_graph._addr_label(addr), style="filled",
        )
        parents = taint_graph._reduced_path_parents.get(addr, [])
        to_process.extend(parents)
        included_keys.add(addr)

    for key, srcs in taint_graph._reduced_path.items():
        for src_key, src in srcs.items():
            if key not in included_keys or src_key not in included_keys:
                continue
            graph.edge(f"{key:x}", f"{src_key:x}", label=src.use_string())

    graph.render()


def get_nodes_and_edges(taint_graph, address_map):
    # print("Setting up source graph")
    # tuples containing source code addr and next addr to check
    open_paths = [(taint_graph._start_addr, taint_graph._start_addr)]
    # No need to analyze an address that has already been analyzed.
    analyzed_addrs = set()
    # use to defines
    edges = defaultdict(list)
    nodes = set()
    while len(open_paths) > 0:
        child, next_ancestor = open_paths.pop()
        if next_ancestor in analyzed_addrs:
            continue
        analyzed_addrs.add(next_ancestor)
        parents = taint_graph._reduced_path_parents.get(next_ancestor, [])
        for parent in parents:
            if address_map.get(parent, None) is not None:
                edges[child].append(parent)
                nodes.add(child)
                nodes.add(parent)
                if parent not in analyzed_addrs:
                    open_paths.append((parent, parent))
            else:
                if parent not in analyzed_addrs:
                    open_paths.append((child, parent))
    return (nodes, edges)


def render_source_graph(taint_graph, address_map, files):
    try:
        from graphviz import Digraph
    except Exception as e:
        return

    (nodes, edges) = get_nodes_and_edges(taint_graph, address_map)
    graph = Digraph(filename="source_cfg", format="png")
    graph.node_attr["shape"] = "box"
    node_ids = set()
    for n in nodes:
        file, line_num = address_map.get(n, (None, None))
        node_id = f"{file}{line_num}"
        if node_id in node_ids:
            continue
        node_ids.add(node_id)
        # Make the taint graph structure have source information
        values = list(taint_graph._reduced_path[n].values())
        if file not in files:
            src_line = f"0x{n:x}"
        else:
            context = 3
            src_lines = []
            for i in range(line_num - context, line_num + context + 1):
                if i < 0 or i >= len(files[file]):
                    continue
                line = html.escape(files[file][i].strip())
                if i == line_num and len(values) > 0:
                    line = (
                        f"<u>{line} /*val ="
                        f" {html.escape(str(values[0].val))}*/</u>"
                    )
                line += f'<br align="left"/>'
                src_lines.append(line)
            text = "".join(src_lines)
            src_line = f"<{text}>"
        node_id = f"{file}{line_num}"
        if node_id in node_ids:
            continue
        node_ids.add(node_id)
        # Make the taint graph structure have source information
        values = list(taint_graph._reduced_path[n].values())
        if file not in files:
            src_line = f"0x{n:x}"
        else:
            context = 3
            src_lines = []
            for i in range(line_num - context, line_num + context + 1):
                if i < 0 or i >= len(files[file]):
                    continue
                line = html.escape(files[file][i].strip())
                if i == line_num:
                    line = (
                        f"<u>{line} /*val = {html.escape(values[0].val)}*/</u>"
                    )
                line += f'<br align="left"/>'
                src_lines.append(line)
            text = "".join(src_lines)
            src_line = f"<{text}>"
        graph.node(node_id, label=src_line, style="filled")
    for src, dests in edges.items():
        srcfile, srcline_num = address_map.get(src, (None, None))
        for dest in dests:
            destfile, destline_num = address_map.get(dest, (None, None))
            graph.edge(
                f"{destfile}{destline_num}",
                f"{srcfile}{srcline_num}",
                label="",
            )
    graph.render()



# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================

from collections import defaultdict

from zelos import Zelos, IPlugin, CommandLineOption, HookType
import hexdump

CommandLineOption("asan", action="store_true", help="ASAN-like capabilities")


class WriteRecord:
    def __init__(self, ip_addr, mem_addr, value):
        self.ip_addr = ip_addr
        self.mem_addr = mem_addr
        self.value = value

    def __repr__(self):
        return f"(ip 0x{self.ip_addr:x}: {self.value} -> 0x{self.mem_addr:x})"


class AllocInfo:
    def __init__(self, addr, size, inst_addr, desc, is_free=False):
        # Address of the corresponding heap buffer allocation
        self.address = addr
        # Size of the corresponding heap buffer allocation
        self.size = size
        # Address of the corresponding heap buffer allocation inst.
        self.inst_address = inst_addr
        # Short string describing the allocation site,
        #   e.g. "malloc(0x100)"
        self.description = desc
        # True if the corresponding allocation had been free'd
        self.is_free = is_free

        self.writes = defaultdict(list)

    def __str__(self):
        return f"Origin Inst: 0x{self.inst_address:x} {self.description}"

    def _record_write(self, ip_addr: int, mem_addr: int, value: int):
        self.writes[mem_addr].append(WriteRecord(ip_addr, mem_addr, value))

    def get_writes(self, mem_addr: int, size: int):
        """
        Returns all addresses that wrote to this memory region
        """
        accessors = []
        for addr in range(mem_addr, mem_addr + size):
            accessors.extend(self.writes[addr])
        return accessors

    def summarize_buffer(self, memory) -> str:
        """
        Returns a string summarizing the memory buffer.
        """
        memory_buffer = memory.read(self.address, self.size)
        chunk_size = 16
        chunks = [
            memory_buffer[i : i + chunk_size]
            for i in range(0, len(memory_buffer), chunk_size)
        ]
        lines = []
        current_chunk = None
        duplicate_line_count = 0
        for i, chunk in enumerate(chunks):
            if current_chunk == chunk:
                if lines[-1] != "...":
                    lines.append("...")
                duplicate_line_count += 1
                continue
            elif len(lines) > 0 and lines[-1] == "...":
                lines[
                    -1
                ] = f"... omitting {duplicate_line_count} duplicate lines"
                duplicate_line_count = 0

            current_chunk = chunk
            line = hexdump.hexdump(chunk, result="return")
            offset, rest = line.split(":", 1)
            offset = int(offset, 16) + i * chunk_size
            writes = self.get_writes(self.address + offset, chunk_size)
            line = f"{offset:08x}:{rest} {writes}"
            lines.append(line)

        if len(lines) > 20:
            lines = (
                ["First ten lines:"]
                + lines[:10]
                + [f"Last ten lines:"]
                + lines[-10:]
            )

        return "  Buffer Contents:\n" + "\n".join(lines)


class CrashInfo:
    def __init__(
        self,
        reason: str = "",
        operation: str = "",
        inst_address=None,
        mem_address=None,
        mem_access_size=None,
        alloc_info=None,
    ):
        # Short description of the crash problem
        self.reason = reason
        # Either READ, WRITE or the empty string
        self.operation = operation
        # Address of the crashing instruction
        self.inst_address = inst_address
        # Address of the memory location causing the crash
        self.mem_address = mem_address
        # Number of bytes read starting at the mem_address
        self.mem_access_size = mem_access_size
        # Information about the buffer of origin for the crash
        self.alloc_info = alloc_info

    @property
    def exploitability(self):
        """
        An estimate on whether the crash is exploitable
        """
        reason = self.reason
        operation = self.operation
        addr = self.mem_address

        near_zero = False
        if addr > 0 and addr < 0x1000:
            near_zero = True

        if reason in ["double-free", "bad-free"]:
            return "EXPLOITABLE"
        if reason == "heap-use-after-free":
            return "EXPLOITABLE"
        if reason == "heap-overflow":
            if operation == "READ":
                if near_zero:
                    return "PROBABLY_NOT_EXPLOITABLE"
                else:
                    return "UNKNOWN"
            if operation == "WRITE":
                if near_zero:
                    return "PROBABLY_EXPLOITABLE"
                else:
                    return "EXPLOITABLE"
        return "UNKNOWN"

    def get_summary(self, memory):
        s = f"\nCrash Summary:\n"
        s += f"  Exploitable: {self.exploitability}\n"
        s += f"  Reason: {self.reason}\n"
        s += f"  Crashing Instruction: 0x{self.inst_address:x}\n"
        if self.mem_address is not None:
            s += f"  Crashing Access: {self.operation} 0x{self.mem_address:x}"
        if self.alloc_info is not None:
            if self.mem_address:
                s += f" (buf + 0x{(self.mem_address-self.alloc_info.address):x})"
                s += f" size: {self.mem_access_size:x} byte(s)\n"
            else:
                s += "\n"
            s += "  " + str(self.alloc_info) + "\n"
            s += self.alloc_info.summarize_buffer(memory)
        else:
            s += "\n"
        return s


class Asan(IPlugin):
    """
        Implements heap sanitization similar to ASAN or libdislocator.

        Specifically:
            - Creates a GUARD_SIZE region of protected memory:
                - Immediately following malloc'd buffers
    """

    def __init__(self, zelos: Zelos):
        super().__init__(zelos)

        self._allocs = {}
        self._guard_size = 0x10
        self._crash_info = None

        if not zelos.config.asan:
            return

        # Used to correct invalid crashing address for runs with
        # INST.EXEC hook.
        self._inst_hook_triggered = False
        self._add_hooks()

    def _add_hooks(self):
        """ Add linux memory allocator hooks and generic invalid
            memory access hook.
        """
        hooks = {
            "malloc": ([("void*", "ptr")], self._malloc),
            "calloc": ([("size_t", "num"), ("size_t", "size")], self._calloc),
            "realloc": (
                [("void*", "ptr"), ("size_t", "new_size")],
                self._realloc,
            ),
            "free": ([("void*", "ptr")], self._free),
        }
        for fn_name, fn_hook in hooks.items():
            self.zelos.internal_engine.hook_manager.register_func_hook(
                fn_name, fn_hook[1]
            )

        self.zelos.hook_memory(HookType.MEMORY.INVALID, self._invalid_hook)

    def _invalid_hook(self, zelos, access, address, size, value):
        """ Hook invoked any time an invalid memory access is triggered.
        """
        if self._crash_info is None:
            operation = "READ"
            if access == 20:
                operation = "WRITE"

            self._crash_info = CrashInfo(
                reason="unknown-crash",
                operation=operation,
                inst_address=self.zelos.thread.getIP(),
                mem_address=address,
                mem_access_size=size,
                alloc_info=None,
            )

            self.logger.warning(self._crash_info.get_summary(zelos.memory))

    @property
    def asan_guard_triggered(self) -> bool:
        return self.get_crash_alloc_info() is not None

    def get_crash_alloc_info(self) -> AllocInfo:
        if self._crash_info is None:
            return None
        return self._crash_info.alloc_info

    def set_inst_hook(self):
        def inst_hook(z, addr, size):
            pass

        return self.zelos.hook_execution(HookType.EXEC.INST, inst_hook)

    def _add_memory_guard(self, start: int, end: int, alloc_info: AllocInfo):
        def guard_access(zelos, access, addr, size, value):
            if addr + size <= start:
                return
            if not self._inst_hook_triggered:
                # If you are running zelos without an EXEC.INST hook,
                # the crash at the memory guard page will not return
                # the proper instruction address. We set the
                # instruction hook and run again to get the correct
                # address.
                zelos.internal_engine.scheduler.stop_and_exec(
                    "pre-crash inst hook", self.set_inst_hook
                )
                self._inst_hook_triggered = True
                return

            if alloc_info.is_free:
                reason = "heap-use-after-free"
            else:
                reason = "heap-overflow"
            operation = "READ"
            if access == 22:
                operation = "WRITE"

            self._crash_info = CrashInfo(
                reason=reason,
                operation=operation,
                inst_address=self.zelos.thread.getIP(),
                mem_address=addr,
                mem_access_size=size,
                alloc_info=alloc_info,
            )

            self.logger.warning(self._crash_info.get_summary(zelos.memory))

            def crash():
                # Now make the underlying page a guard page so zelos
                # will fault and end execution.
                zelos.memory._memory.protect(
                    start, max(0x1000, end - start), 0
                )

            zelos.internal_engine.scheduler.stop_and_exec("Crash", crash)

        return self.zelos.hook_memory(
            HookType.MEMORY.VALID,
            guard_access,
            mem_low=start - 7,
            mem_high=end,
            end_condition=lambda: self.asan_guard_triggered,
        )

    def _safe_alloc(self, size: int, desc: str = "") -> int:
        """
        Allocates memory and ensures that a crash will happen if memory
        is written outside of the boundaries.
        """
        addr = self.zelos.memory._memory.heap.alloc(
            size + self._guard_size, name="safe_malloc"
        )
        alloc_info = AllocInfo(addr, size, self.zelos.thread.getIP(), desc)
        self._record_writes(alloc_info)
        high_hook = self._add_memory_guard(
            addr + size, addr + size + self._guard_size - 1, alloc_info
        )
        self._allocs[addr] = (size, high_hook)
        return addr

    def _record_writes(self, alloc_info: AllocInfo):
        def record_write(zelos, access, addr, size, value):
            alloc_info._record_write(
                zelos.thread.getIP(),
                addr,
                zelos.memory.pack(value, bytes=size),
            )

        return self.zelos.hook_memory(
            HookType.MEMORY.WRITE,
            record_write,
            mem_low=alloc_info.address,
            mem_high=alloc_info.address + alloc_info.size,
        )

    def _handle_return(self, retval: int):
        if retval is not None:
            self.zelos.internal_engine.kernel.set_return_value(retval)
            dfa = self.zelos.plugins.dataflow
            if dfa.dataflow_enabled:
                dfa.trace.add_define(
                    self.zelos.internal_engine.kernel._REG_RETURN,
                    retval,
                    label_override="Memory API",
                )
        thread = self.zelos.thread
        # FIXME: this line only works on architectures with stack-based
        #   return addresses.
        retaddr = thread.popstack()

        def set_ip():
            thread.setIP(retaddr)

        self.zelos.internal_engine.scheduler.stop_and_exec(
            "syscall_ip_change", set_ip
        )

    def _get_args(self, args):
        return self.zelos.internal_engine.kernel.get_args(args)

    def _malloc(self, zelos):
        """ Add a GUARD_SIZE guard immediately following the buffer.
        """
        args = self._get_args([("int", "size")])
        retval = self._safe_alloc(args.size, f"malloc(size=0x{args.size:x})")
        self._handle_return(retval)

    def _calloc(self, zelos):
        """ Add a GUARD_SIZE guard immediately following the buffer.
        """
        args = self._get_args([("size_t", "num"), ("size_t", "size")])
        size = args.num * args.size
        if size <= 0:
            self._handle_return(0)
        retval = self._safe_alloc(
            size, f"calloc(num=0x{args.num:x}, size=0x{args.size:x})"
        )
        self.zelos.memory.write(retval, b"\x00" * size)
        self._handle_return(retval)

    def _realloc(self, zelos):
        """ Add a GUARD_SIZE guard immediately following the buffer.
        """
        args = self._get_args([("void*", "ptr"), ("size_t", "new_size")])
        if args.ptr == 0:
            retval = self._safe_alloc(
                args.new_size,
                f"realloc(ptr=0x{args.ptr:x}, new_size=0x{args.new_size:x})",
            )
        else:
            retval = self._safe_alloc(
                args.new_size,
                f"realloc(ptr=0x{args.ptr:x}, new_size=0x{args.new_size:x})",
            )
        self._handle_return(retval)

    def _free(self, zelos):
        """ Add guard in the entirety of the free'd buffer space.
        """
        args = self._get_args([("void*", "ptr")])
        if args.ptr == 0:
            return

        # TODO: double-free and bad-free checks
        # self.crash_reason = "double-free"
        # self.crash_reason = "bad-free"
        # if so, end execution

        # Delete the previous GUARD_SIZE guard
        size, high_hook = self._allocs.pop(args.ptr)
        self.zelos.delete_hook(high_hook)
        # Add guard in the free'd space
        alloc_info = AllocInfo(
            args.ptr,
            size,
            self.zelos.thread.getIP(),
            f"free(f{args.ptr:x})",
            is_free=True,
        )
        self._add_memory_guard(
            args.ptr - self._guard_size,
            args.ptr + size + self._guard_size - 1,
            alloc_info,
        )
        self._handle_return(None)



# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================

from zelos.exceptions import MemoryReadUnmapped
from typing import List, Dict
from collections import defaultdict

import math
import re


class TaintNode:
    """
    TODO
    """

    def __init__(self, taint_graph, idx, use, val):
        self.idx = idx
        self.use = use
        self.val = val
        self._taint_graph = taint_graph
        try:
            potential_addr = int(val, 0)
            self._taint_graph._dataflow.zelos.memory.read_int(potential_addr)
            self.val = "*" + self.val
        except Exception:
            # Not a valid memory address
            pass

    def addr(self) -> int:
        return self._taint_graph._dataflow.trace._trace[self.idx]

    def use_string(self) -> str:
        if isinstance(self.use, int):
            return f"Use: 0x{self.use:x} Value: {self.val}"
        else:
            return f"Use: {self.use} Value: {self.val}"

    def __str__(self):
        return f"(Addr: 0x{self.addr():x} {self.use_string()})"

    def __repr__(self):
        return self.__str__()


class ReducedTaintNode:
    """
    TODO
    """

    def __init__(self, taint_graph, taint_node):
        self._taint_graph = taint_graph
        self._taint_nodes = [taint_node]

    @property
    def idx(self):
        return self._taint_nodes[0].idx

    @property
    def use(self):
        return self._taint_nodes[0].use

    @property
    def val(self):
        return self._taint_nodes[0].val

    def addr(self) -> int:
        return self._taint_graph._dataflow.trace._trace[self.idx]

    def use_string(self) -> str:
        if isinstance(self.use, int):
            return (
                f"Use: 0x{self.use:x} Value: {self.val} Iter:"
                f" {len(self._taint_nodes)}"
            )
        else:
            return (
                f"Use: {self.use} Value: {self.val} Iter:"
                f" {len(self._taint_nodes)}"
            )

    def add_node(self, taint_node):
        self._taint_nodes.append(taint_node)

    def __str__(self):
        return f"(Addr: 0x{self.addr():x} {self.use_string()})"

    def __repr__(self):
        return self.__str__()


class TaintGraph:
    """
    TODO
    """

    def __init__(self, dataflow, start_addr, crashing_address=None):
        self.logger = dataflow.logger
        self._dataflow = dataflow

        "Keys are the index in the trace"
        self._path: Dict[int, Dict[int, TaintNode]] = defaultdict(dict)

        self._path_parents = defaultdict(list)

        """Keys are the address. This is an optimization to
        reduce the size of the graph."""
        self._reduced_path: Dict[
            int, Dict[int, ReducedTaintNode]
        ] = defaultdict(dict)

        self._reduced_path_parents = defaultdict(set)
        self._start_addr = start_addr
        self._assembly = defaultdict(str)
        self._addr2func = {}

        self._dwarf_data = None
        self.reverse_taint(start_addr, crashing_address)

    def __str__(self):
        s = ""
        count = 100
        if len(self._reduced_path) > 0:
            for addr, defines in self._reduced_path.items():
                s += f"0x{addr:x} {defines}\n"
                s += f"  assembly: {self._assembly[addr]}\n"
                funcname = self._addr2func.get(addr, None)
                if funcname is not None:
                    s += f"  func: {funcname}\n"
                srccode = self.get_source(addr)
                if srccode is not None:
                    s += f"  src: {srccode}\n"
                count -= 1
                if count == 0:
                    break
            return s

        for idx, defines in self._path.items():
            addr = self._dataflow.trace._trace[idx]
            s += f"0x{addr:x} {defines}\n"
        return s

    @property
    def reduced_path(self):
        return self._reduced_path

    def get_source(self, addr: int):
        if self._dwarf_data is None:
            return None
        return self._dwarf_data.get_source(addr)

    def get_assembly_from_source(self, file, line_num):
        address_map = self._dwarf_data._address_map
        (addr_low, addr_high) = address_map.get_addr_range_from_source(
            file, line_num
        )
        if addr_low is None or addr_high is None:
            return None
        return self.get_assembly_for_range(addr_low, addr_high)

    def _reduce(self):
        """
        Creates the reduced path which ensures that uses the address as
        the key instead of the index
        """
        if len(self._reduced_path) > 0:
            return  # Already reduced
        for idx, taint_nodes in self._path.items():
            addr = self._dataflow.trace._trace[idx]
            for taint_node in taint_nodes.values():
                node_addr = self._dataflow.trace._trace[taint_node.idx]
                if node_addr in self._reduced_path[addr]:
                    self._reduced_path[addr][node_addr].add_node(taint_node)
                else:
                    self._reduced_path[addr][node_addr] = ReducedTaintNode(
                        self, taint_node
                    )
                self._reduced_path_parents[node_addr].add(addr)

            # Add information about the assembly instructions
            if addr not in self._assembly:
                inst_strings = self.get_assembly_for_range(addr, addr + 20)
                self._assembly[addr] = inst_strings[addr]

        # Filter out adjacent push-pop pairs
        for addr, taint_nodes in list(self._reduced_path.items()):
            if " push " not in self._assembly.get(addr, ""):
                continue
            if len(taint_nodes) != 1:
                continue
            child_addr = list(taint_nodes.keys())[0]
            node = taint_nodes[child_addr]
            if " pop " not in self._assembly[child_addr]:
                continue
            grand_children = {
                k: v
                for k, v in self._reduced_path[child_addr].items()
                if "push" not in self._assembly[k]
            }
            grand_parents = [
                x
                for x in self._reduced_path_parents[addr]
                if "pop" not in self._assembly[x]
            ]
            # print("Beginning Deletion")
            # for gp in grand_parents:
            #     print(f"  grand parent 0x{gp:x} {self._assembly[gp]}")
            #     print(f"    children: {self._reduced_path[gp]}")
            # print(f"  parent: {addr:x} {self._assembly[addr]}")
            # print(f"  child: {child_addr:x} {self._assembly[child_addr]}")
            # for gc_addr in grand_children.keys():
            #     print(
            #         f"  grand_children {gc_addr:x} {self._assembly[gc_addr]}"
            #     )

            for gp in grand_parents:
                del self._reduced_path[gp][addr]
            del self._reduced_path[addr]
            del self._reduced_path[child_addr]
            del self._reduced_path_parents[addr]
            del self._reduced_path_parents[child_addr]

            # print(
            #     f"  Deleted push-pop pair {self._assembly[addr]} and {self._assembly[child_addr]}"
            # )

            for grand_child_addr, grand_child in grand_children.items():
                # print(f"  adding grand child 0x{grand_child_addr:x}")
                self._reduced_path_parents[grand_child_addr].remove(child_addr)
                self._reduced_path_parents[grand_child_addr].update(
                    grand_parents
                )
                for gp in grand_parents:
                    self._reduced_path[gp][grand_child_addr] = grand_child

            # print("After:")

            # for gp in grand_parents:
            #     print(" ", [hex(k) for k in self._reduced_path[gp].keys()])
            # for gc in grand_children.keys():
            #     print(f" Grand child {gc:x}",)
            #     print(
            #         f"    grand parents: {[hex(x) for x in self._reduced_path_parents[gc]]}"
            #     )

    def get_assembly_for_range(self, addr_low, addr_high):
        try:
            code = self._dataflow.memory.read(addr_low, addr_high - addr_low)
        except MemoryReadUnmapped:
            print(f"Error trying to read 0x{addr_low:x}-0x{addr_high:x}")
            return {}
        if self._dataflow.zelos.config.link_ida is not None:
            try:
                ida_disasm = self._dataflow.zelos.plugins.ida.idc.GetDisasm(
                    addr_low
                )
                if ida_disasm is not None and ida_disasm != "":
                    self._annotate_variables_in_path(addr_low, ida_disasm)
                    return {
                        addr_low: (
                            f"0x{addr_low:x}: {ida_disasm} ;"
                            f" {self._dataflow.trace._addr_comment.get(addr_low, '')}"
                        )
                    }
            except Exception as e:
                print("Ida address exception: ", e)
                # raise e
                pass
        inst_list = list(
            self._dataflow.zelos.internal_engine.cs.disasm(code, addr_low)
        )
        return {
            i.address: (
                f"0x{i.address:x}: {i.mnemonic} {i.op_str} ;"
                f" {self._dataflow.trace._addr_comment.get(i.address, '')}"
            )
            for i in inst_list
        }

    def _annotate_variables_in_path(self, addr: int, ida_disasm: str):
        if ida_disasm is None:
            return
        if "mov" not in ida_disasm and "lea" not in ida_disasm:
            return

        cmd, args = ida_disasm.split(maxsplit=1)
        if cmd not in ["mov", "lea"]:
            return
        dest, src = args.split(",", maxsplit=1)
        var_name = None
        if "[" in dest and "[" not in src:
            var_name = self._get_ida_variable(dest)
        if "[" in src and "[" not in dest:
            var_name = self._get_ida_variable(src)
        if var_name is None:
            return
        if cmd == "lea":
            var_name = "&" + var_name
        for reduced_nodes in self._reduced_path[addr].values():
            for node in reduced_nodes._taint_nodes:
                node.val = f"{var_name}={node.val}"

    def _get_ida_variable(self, s):
        """
        Get the variable name from Ida disassembly if it exists.
        """
        match = re.search(r"\[.bp\+(\w+)\]", s)
        if match is not None:
            return match.group(1)
        match = re.search(
            r"\[.sp\+[0-9A-Fa-f]+h\+([A-Za-z][A-Za-z0-9.]*)\]", s
        )
        if match is None:
            return None
        return match.group(1)

    def reverse_taint(self, start_addr: int, crashing_addr: int):
        """
        Returns a list of addresses.
        Each address also contains what address tainted them and why.
        """

        start_idx = self._dataflow.trace._last_idx_of_address(start_addr)
        self._path[start_idx][-1] = TaintNode(self, -1, -1, -1)
        indices_to_analyze = []
        has_been_analyzed = set()

        self._add_uses(
            start_idx, indices_to_analyze, restrict_to_value=crashing_addr
        )

        while len(indices_to_analyze) > 0:
            current_idx = indices_to_analyze.pop()

            # Print how long the path on occasion
            self._maybe_log(current_idx, indices_to_analyze)
            self._add_uses(current_idx, indices_to_analyze)

        self._reduce()

    def _add_uses(
        self,
        current_idx: int,
        indices_to_analyze: List[int],
        restrict_to_value=None,
    ):
        reg_uses = self._dataflow.ud_chain(current_idx)
        for use, define_idxs in reg_uses.items():
            if use in [
                "ebp",
                "esp",
                "rbp",
                "rsp",
                "cc_src",
                "cc_src2",
                "cc_dst",
                "cc_op",
                "eq",
                "ne",
                "ltu",
                "leu",
                "gtu",
                "geu",
            ]:
                continue
            for define_idx in define_idxs:
                defines = self._dataflow.trace.get_defines(define_idx)
                val = defines[use]
                if (
                    restrict_to_value is not None
                    and hex(restrict_to_value) != val
                ):
                    continue
                if define_idx not in self._path:
                    indices_to_analyze.append(define_idx)
                self.create_taint_node(
                    define_idx, current_idx, use, val,
                )

        mem_uses = self._dataflow.trace.get_uses(current_idx)
        for use, define_idx in mem_uses.items():
            if define_idx is None:
                inst_addr = self._dataflow.trace._trace[current_idx]
                self.logger.warning(
                    "Memory use has not been defined "
                    f"inst addr: {inst_addr:x} mem_addr: {use:x}"
                )
                continue
            defines = self._dataflow.trace.get_defines(define_idx)
            val = defines[use]
            if restrict_to_value is not None and hex(restrict_to_value) != val:
                continue
            if define_idx not in self._path:
                indices_to_analyze.append(define_idx)
            self.create_taint_node(
                define_idx, current_idx, use, val,
            )

    def _maybe_log(self, current_idx: int, indices_to_analyze: List[int]):
        path_len = len(self._path)
        order_of_magnitude = 10 ** math.floor(math.log10(path_len))
        if path_len % max(100, order_of_magnitude) == 0:
            self.logger.info(
                f"Path size: {path_len}, Analyzing idx {current_idx}"
                f" left:{len(indices_to_analyze)}"
            )

    def create_taint_node(self, define_idx, current_idx, use, val):
        self._path[define_idx][current_idx] = TaintNode(
            self, current_idx, use, val
        )
        self._path_parents[current_idx].append(define_idx)

    def _addr_label(self, key: int):
        if len(self._reduced_path) > 0:
            address = key
        else:
            address = self._dataflow.trace._trace[key]

        if self._dataflow.zelos.config.link_ida is not None:
            name = self._dataflow._get_ida_func_name(address)
            func_addr = self._dataflow.names.get(name, None)
        else:
            func_addr = None

        s = self._assembly[address]

        region = self._dataflow.zelos.memory.get_region(address)

        label_override = self._dataflow.trace.label_overrides.get(
            address, None
        )
        if label_override is not None:
            label = label_override
        elif func_addr is None:
            label = f"0x{address:x} {region.name} {region.module_name}"
        else:
            label = f"{name}+0x{address-func_addr:x}"
        s = f"{label}\n{s}\n"
        source_code = self.get_source(address)
        if source_code is not None:
            s += source_code + "\n"
        return s



# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================

import subprocess
import pickle
import base64
import os
import os.path as path
import tempfile

from zelos import IPlugin, CommandLineOption

CommandLineOption(
    "link_ida",
    type=str,
    default=None,
    help="Absolute path to an instance of IDA Pro",
)


class Ida(IPlugin):
    def __init__(self, zelos):
        super().__init__(zelos)
        self.initialized = False
        if zelos.config.link_ida is None:
            return

        ida_path = zelos.config.link_ida
        if not os.path.exists(ida_path):
            self.logger.error(f"Cannot resolve path to IDA executable.")
            return

        server_path = path.join(
            path.dirname(path.abspath(__file__)), "ida_server.py"
        )
        self.temp_dir = tempfile.TemporaryDirectory()

        ida_input_path = path.join(self.temp_dir.name, "ida_input")
        os.mkfifo(ida_input_path)

        ida_output_path = path.join(self.temp_dir.name, "ida_output")
        os.mkfifo(ida_output_path)

        p = subprocess.Popen(
            [
                ida_path,
                "-c",
                "-A",
                f'-S""{server_path} {ida_input_path} {ida_output_path}"" ',
                zelos.target_binary_path,
            ]
        )
        self.ida_input = open(ida_input_path, "w")
        self.ida_output = open(ida_output_path, "rb")

        def cleanup_proc():
            p.terminate()
            p.wait()
            self.temp_dir.cleanup()

        zelos.hook_close(cleanup_proc)

        self.initialized = True
        self.rebased = False

    def _auto_rebase(self):
        if self.rebased:
            return
        zelos = self.zelos
        zelos_base = zelos.internal_engine.files.get_module_base_by_name(
            zelos.target_binary_path
        )
        ida_base = self._raw_exec("idaapi.get_imagebase()")
        print(f"Adjusting imagebase from: {ida_base:x} to {zelos_base}")
        delta = zelos_base - ida_base

        # 8 == ida_segment.MSF_FIXONCE
        self._raw_exec(f"ida_segment.rebase_program({delta}, 8)")
        self.rebased = True

    def __getattr__(self, module_name):
        if self.initialized:
            return FauxModule(self, f"{module_name}")
        return super().__getattr__(self, module_name)

    @property
    def api(self):
        return FauxModule(self, f"idaapi")

    @property
    def utils(self):
        return FauxModule(self, f"idautils")

    @property
    def idc(self):
        return FauxModule(self, f"idc")

    def _exec(self, cmd: str) -> str:
        if not self.rebased:
            self._auto_rebase()
        return self._raw_exec(cmd)


    def _raw_exec(self, cmd: str) -> str:
        """
        Sends a command to IDA for execution
        """
        self.ida_input.write(f"{cmd}\n")
        self.ida_input.flush()
        data = self.ida_output.readline()
        data = base64.b64decode(data.strip())
        return pickle.loads(data, encoding="latin1")


class FauxModule:
    """
    This is a class used to identify what methods are being called on
    ida so that they can be passed as a string to the IdaServer
    """

    def __init__(self, ida_server_plugin, field_name):
        self._ida_server_plugin = ida_server_plugin
        self.name = field_name

    def __getattr__(self, field_name):
        return FauxModule(
            self._ida_server_plugin, self.name + "." + field_name
        )

    def _stringify(self, val):
        if type(val) == str:
            return '"' + val + '"'
        return str(val)

    def __call__(self, *args, **kwargs):
        arguments = [self._stringify(x) for x in args] + [
            f"{k}={self._stringify(v)}" for k, v in kwargs
        ]
        return self._ida_server_plugin._exec(
            self.name + "(" + ",".join(arguments) + ")"
        )



# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================

from collections import defaultdict
import os
from pathlib import Path
from typing import Optional

from elftools.dwarf.descriptions import describe_form_class
from elftools.elf.elffile import ELFFile


class AddressEntry:
    def __init__(self, low, high, file, line_num):
        self.low = low
        self.high = high
        self.file = file
        self.line_num = line_num

    def __str__(self):
        return f"(0x{self.low:x}-0x{self.high:x}: {self.file}:{self.line_num})"


class AddressMap:
    def __init__(self):
        self._address_low_high = []
        self._cache = {}
        self._source2addr_range = {}

    def __str__(self):
        return [entry.__str__() for entry in self._address_low_high].__str__()

    def files(self):
        return {entry.file for entry in self._address_low_high}

    def get_addr_range_from_source(self, file, line_num):
        return self._source2addr_range.get(f"{file}{line_num}", (None, None))

    def _add_cache(self, addr, file, line):
        self._cache[addr] = (file, line - 1)

    def add(self, low, high, file, line):
        file = file.decode()
        self._add_cache(low, file, line)
        self._source2addr_range[f"{file}{line}"] = (low, high)
        for addr in range(low, high):
            self._add_cache(addr, file, line)

        entry = AddressEntry(low, high, file, line)
        self._address_low_high.append(entry)
        self._address_low_high.sort(key=lambda x: x.low)

    def get(self, addr: int, default=(None, None)):
        cached_val = self._cache.get(addr, None)
        if cached_val is not None:
            return cached_val

        idx = self.binary_search(addr)
        if idx is not None:
            entry = self._address_low_high[idx]
            retval = (entry.file, entry.line_num - 1)
            self._cache[addr] = retval
            return retval

        return default

    def _attach_src_to_external_addrs(self, trace, dwarf_data):
        """
        Certain addresses in the trace are not associated with source
        code because they are in external modules. We associate those
        addresses with source code from the last line that occurred
        before moving to the external module.
        """
        # External modules can call each other. This means that the same
        # address could have multiple candidates for what address should
        # be associated with them.
        file_lines = dwarf_data._file_lines

        current_source = None
        current_file = None
        current_line = None
        smeared = {}
        for addr in trace:
            file, line_num = self.get(addr)

            if file in file_lines and addr not in smeared:
                dwarf_data._addr2source[addr] = file_lines[file][line_num]
                current_source = file_lines[file][line_num]
                current_file = file
                current_line = line_num
            elif current_source is not None:
                dwarf_data._addr2source[addr] = "within " + current_source
                self._add_cache(addr, current_file, current_line + 1)
                smeared[addr] = current_source

    def binary_search(self, x):
        low = 0
        mid = 0
        high = len(self._address_low_high) - 1
        while low <= high:
            mid = (high + low) // 2
            entry = self._address_low_high[mid]
            if entry.low <= x < entry.high:
                return mid
            if entry.low == entry.high and entry.low == x:
                return mid

            # Check if x is present at mid
            if entry.low < x:
                low = mid + 1

            # If x is greater, ignore left half
            elif entry.low > x:
                high = mid - 1

        # If we reach here, then the element was not present
        return None


class DwarfData:
    """
    Class to parse and hold relevant Dwarf information from binary
    """

    def __init__(self, binary_path, source_path, rebased_module_base=None):
        self._binary_path = binary_path
        self._source_path = source_path
        self._rebased_module_base = rebased_module_base
        self._fp = None
        self._elf = None
        self._dwarfinfo = None
        self._offset = 0
        self._file_lines = {}
        self._file_to_syspath = {}
        self._addr2source = {}
        self._address_map = None
        self._function_map = {}

        self.__load_elf_and_dwarf(binary_path)
        self.__calculate_offset(rebased_module_base)
        self.__set_address_map()
        self.__setup_file_lines()
        self.__build_function_map()

    def __del__(self):
        if hasattr(self._fp, "close"):
            self._fp.close()

    def get_source(self, addr: int) -> Optional[str]:
        return self._addr2source.get(addr, None)

    def get_function_info(self, taint_graph):
        function_map = self._function_map
        addr2func = {}
        for addr in taint_graph._reduced_path.keys():
            for f, ranges in function_map.items():
                for r in ranges:
                    (low, high) = r
                    if low <= addr <= high:
                        addr2func[addr] = f
        return addr2func

    def attach_src_to_external_addrs(self, trace):
        if self._address_map is None:
            return
        self._address_map._attach_src_to_external_addrs(trace, self)

    def __get_elf_module_base(self, elf):
        # This function DOES NOT WORK for static binaries.
        # It gets lucky on binaries where the main module is loaded at 0,
        # which is what this function normally returns. Have to find a way
        # to get the desired load address of binaries. Maybe one way to get
        # around is always return 0 for dynamic binaries.
        segment_addrs = [s.header.p_vaddr for s in elf.iter_segments()]
        # print(f"Segment_addrs: {segment_addrs}")
        return min(segment_addrs)

    def __load_elf_and_dwarf(self, binary_path):
        if os.path.exists(binary_path) and os.path.isfile(binary_path):
            self._fp = open(binary_path, "rb")
            self._elf = ELFFile(self._fp)
            try:
                self._dwarfinfo = self._elf.get_dwarf_info()
                self._elf_module_base = self.__get_elf_module_base(self._elf)
            except:
                pass

    def __calculate_offset(self, rebased_module_base):
        if rebased_module_base is None:
            return
        symbols_module_base = self._elf_module_base
        offset = rebased_module_base - symbols_module_base
        # TODO: TEMP for static binaries
        if offset != 0x10000:
            offset = 0
        # print("Got offset: ", offset)
        self._offset = offset

    def __set_address_map(self):
        if self._dwarfinfo is None:
            return
        dwarfinfo = self._dwarfinfo
        offset = self._offset
        address_map = AddressMap()
        # Go over all the line programs in the DWARF information, looking for
        # one that describes the given address.
        for CU in dwarfinfo.iter_CUs():
            # First, look at line programs to find the file/line for the addr
            lineprog = dwarfinfo.line_program_for_CU(CU)
            prevstate = None
            for entry in lineprog.get_entries():
                # We're interested in entries where a new state is assigned
                if entry.state is None:
                    continue
                if entry.state.end_sequence:
                    # if the line number sequence ends, clear prevstate.
                    prevstate = None
                    continue
                # Looking for a range of addresses in two consecutive states
                # that contain the required address.
                if prevstate:
                    filename = lineprog["file_entry"][prevstate.file - 1].name
                    line = prevstate.line
                    address_map.add(
                        prevstate.address + offset,
                        entry.state.address + offset,
                        filename,
                        line,
                    )
                prevstate = entry.state
        self._address_map = address_map

    def __resolve_filepath(self, path):
        """
        First checks the current directory for the path.
        Then checks recursively in the source code folder for the path
        """
        if os.path.exists(path):
            return path
        matching_files = [path for path in Path(self._source_path).rglob(path)]
        if len(matching_files) == 0:
            # print(
            #     f"Could not find source code file {path} within"
            #     f" {source_code_path}"
            # )
            return None
        if len(matching_files) > 1:
            # print(
            #     f"There is more than one matching file for {path}:"
            #     f" {matching_files}. Picking {matching_files[0]}"
            # )
            pass
        return matching_files[0]

    def __setup_file_lines(self):
        for filename in self._address_map.files():
            resolved_filename = self.__resolve_filepath(filename)
            if resolved_filename is None:
                continue
            f = open(resolved_filename, "r")
            self._file_to_syspath[filename] = resolved_filename
            # Keep this keyed by the original file name so that later on we
            # can find these lines
            self._file_lines[filename] = list(f.readlines())

    def __build_function_map(self):
        """
        Builds the mapping of function names to
        list of tuples: [(low, high), ...]
        """
        dwarfinfo = self._dwarfinfo
        offset = self._offset
        if dwarfinfo is None:
            return
        functions = defaultdict(list)
        entries = []
        for CU in dwarfinfo.iter_CUs():
            for DIE in CU.iter_DIEs():
                die_info = self.__handle_DIE(DIE)
                if die_info is not None:
                    entries.append(die_info)
        for entry in entries:
            func_name = entry["name"]
            low, high = entry["range"]
            functions[func_name].append((low + offset, high + offset))
        self._function_map = functions

    def __handle_DIE(self, DIE):
        def __extract_value(attr, key, default):
            if key in attr.keys():
                return attr[key].value
            return default

        # we are interested in two things: name and address range
        tag = DIE.tag
        attr = DIE.attributes
        # ignore compile unit info
        if tag == "DW_TAG_compile_unit":
            return None
        # check for low_pc
        lowpc = __extract_value(attr, "DW_AT_low_pc", -1)
        # we don't care if DIE holds no address
        if lowpc == -1:
            return None
        elif "DW_AT_high_pc" in attr.keys():
            highpc_attr = attr["DW_AT_high_pc"]
            highpc_attr_class = describe_form_class(highpc_attr.form)
            if highpc_attr_class == "address":
                highpc = highpc_attr.value
            elif highpc_attr_class == "constant":
                highpc = lowpc + highpc_attr.value
        else:
            highpc = lowpc
        # recursive search for name
        current_die = DIE
        while True:
            name = __extract_value(attr, "DW_AT_name", b"")
            if name and current_die.tag == "DW_TAG_subprogram":
                return {"name": name.decode(), "range": (lowpc, highpc)}
            origin = __extract_value(attr, "DW_AT_abstract_origin", -1)
            if origin == -1:
                break
            current_die = current_die.get_DIE_from_attribute(
                "DW_AT_abstract_origin"
            )
            attr = current_die.attributes
        return None



# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================

from typing import List, Set, Optional

import ctypes

ARITHMETIC_TCG_INSTS = [
    "add",
    "sub",
    "shr",
    "shl",
    "sar",
    "rotl",
    "rotr",
    "and",
    "nand",
    "xor",
    "or",
    "rem",
    "div",
    "mul",
]
IGNORED_TCG_INSTS = [
    "goto_tb",
    "exit_tb",
    "set_label",
    "end",
    "nop",
    "deposit",  # Not sure what this one does
    "discard",
]


def _decode_env_registers(offset):
    """
    Converts registers that are of the form env_$
    """
    reg = {
        "$0x328": "xmm1",
        "$0x330": "xmm1",
        "$0x338": "xmm2",
        "$0x340": "xmm2",
        "$0x348": "xmm3",
        "$0x350": "xmm3",
        "$0x358": "xmm4",
        "$0x360": "xmm4",
        "$0x368": "xmm5",
        "$0x370": "xmm5",
    }.get(offset, None)
    if reg is None:
        return set()
    return set([reg])


class TCGInst:
    """
    Represents Qemu TCG instruction-level uses and defines.
    """

    def __init__(self, inst_str: str):
        """
        Example `inst_str`s:
            1) "movi_i64 tmp3,$0x18f8d"
            2) "st_i64 tmp3,env,$0x80"
        """
        args = ""
        if " " in inst_str:
            self.name, args = inst_str.split(" ", 1)
        else:
            self.name = inst_str
        self.args = args.split(",")

    def __str__(self):
        return (
            f"({self.name}, {self.args}, def:{self.defines()},"
            f" use:{self.uses()})"
        )

    def __repr__(self):
        return self.__str__()

    def defines(self) -> Set[str]:
        """
        Returns the name of registers/temporaries that are defined
        by this instruction.
        """
        if self.is_type(["movcond_"]):
            return set([self.args[1]])
        if self.is_type(["st_i64"]):
            if self.args[1] == "env":
                return _decode_env_registers(self.args[2])
        if self.is_type(
            ARITHMETIC_TCG_INSTS
            + ["mov", "qemu_ld_", "ld", "ext", "neg", "not"]
        ):
            return set([self.args[0]])
        if self.is_type(
            IGNORED_TCG_INSTS + ["call", "br", "brcond", "setcond"]
        ):
            return set()
        if self.is_type(["qemu_st_", "st32_i64"]):
            # Do something with these?
            return set()
        print(
            "[TCGInst] `defines` not parsed:", self.name, ",".join(self.args)
        )
        return set()

    def uses(self) -> Set[str]:
        """
        Returns that name of registers/temporaries that are used
        by this instruction
        """
        if self.is_type(["ld_i64"]):
            if self.args[1] == "env":
                return _decode_env_registers(self.args[2])
        if self.is_type(["movcond"]):
            return set(self.args[2:])
        if self.is_type(["neg", "not"]):
            return set([self.args[0]])
        if self.is_type(["qemu_st_", "st"]):
            return set(self.args[0:2])
        if self.is_type(["call"]):
            return set(self.args[3:])
        if self.is_type(["mov_", "qemu_ld_", "brcond_", "ld", "ext"]):
            return set([self.args[1]])
        if self.is_type(["setcond_"]):
            return set(self.args[0:3])
        if self.is_type(ARITHMETIC_TCG_INSTS):
            return set(self.args[1:3])
        if self.is_type(IGNORED_TCG_INSTS + ["movi_", "br"]):
            return set()
        print(f"[TCGInst] `uses` not parsed: {self.name}", ",".join(self.args))
        return set()

    def is_type(self, list_of_types: List[str]) -> bool:
        """
        This takes a list of prefixes, and ensures that this inst is
        one of them
        """
        return any([self.name.startswith(t) for t in list_of_types])


class TargetInst:
    """
    Represents a group of TCGinsts that were generated from a single
    target architecture instruction.
    """

    def __init__(self, address, header_insts, tcginst_list):
        self.address = address
        self.header_insts = header_insts
        self.tcg_insts = tcginst_list
        # self._validate_tmps()  # For debugging purposes

    def __str__(self):
        inst_string = "\n  ".join([str(i) for i in self.tcg_insts])
        return (
            f"(Address: 0x{self.address:x}"
            + "\n  "
            + inst_string
            + "\n"
            + f"Defs: {self.defines()} Uses: {self.uses()})"
            + "\n"
        )

    def __repr__(self):
        return self.__str__()

    def uses(self) -> Set[str]:
        """
        Iterates over all the TCGInst's representing this instruction
        and accumulates all their uses.
        """
        use_strings = set()
        for inst in self.tcg_insts:
            use_strings.update(
                [
                    u
                    for u in inst.uses()
                    if not (u.startswith("tmp") or u.startswith("loc"))
                ]
            )
        return use_strings

    def defines(self) -> Set[str]:
        """
        Iterates over all the TCGInst's representing this instruction
        and accumulates all their defines.
        """
        def_strings = set()
        for inst in self.tcg_insts:
            def_strings.update(
                [
                    d
                    for d in inst.defines()
                    if not (d.startswith("tmp") or d.startswith("loc"))
                ]
            )
        return def_strings

    def _validate_tmps(self):
        """
        Ensure that every tmp variable has
          * > 0 def and > 0 uses
          * the def happens before the uses
        """
        defined = set()
        used = set()
        for inst in self.tcg_insts:
            defs = [d for d in inst.defines() if d.startswith("tmp")]
            defined.update(defs)
            uses = [u for u in inst.uses() if u.startswith("tmp")]
            for u in uses:
                assert u in defined, (self, u)
                used.add(u)
        assert used == defined, (self, used, defined)


class TCGParse:
    def __init__(self):
        pass

    def get_tcg(self, zelos):
        uc = zelos.internal_engine.emu._uc
        # TODO: Need to get the number of bytes that were written to
        # the buffer so that we can ensure that we got the full tcg
        # string from unicorn. For now, we just pick a big number and
        # use that.
        size = 1000000
        buffer = ctypes.create_string_buffer(size)
        uc.get_tcg_x86_64(buffer, size)

        insts = buffer.value.decode().split("\n")
        tcg_insts = [TCGInst(i.strip()) for i in insts if i != ""]
        assert tcg_insts[-1].name == "end"
        target_insts = self._split_into_target_insts(tcg_insts)
        return target_insts

    def _split_into_target_insts(self, tcg_list):
        """"""
        target_insts = None
        target_inst_grouping = []
        for i in tcg_list:
            target_inst_grouping.append(i)
            if i.name == "call" and i.args[0] == "uc_tracecode":
                if target_insts is None:
                    target_insts = []
                    target_inst_grouping = target_inst_grouping[-5:]
                else:
                    target_inst_grouping, new_inst_grouping = (
                        target_inst_grouping[:-5],
                        target_inst_grouping[-5:],
                    )
                    target_inst = self._make_target_inst(target_inst_grouping)
                    if target_inst is not None:
                        target_insts.append(target_inst)
                    target_inst_grouping = new_inst_grouping
        if len(target_inst_grouping) > 0:
            target_inst = self._make_target_inst(target_inst_grouping)
            if target_inst is not None:
                target_insts.append(target_inst)
        return target_insts

    def _make_target_inst(self, tcg_list) -> Optional[TargetInst]:
        """
        Cleanup tcg_list by removing some patterns that don't seem to
        contribute to the understanding of the original code
        """
        new_tcg_list = []
        i = 0
        while i < len(tcg_list):
            inst = tcg_list[i]
            if inst.name in ["nopn", "nop"]:
                i += 1
                continue
            new_tcg_list.append(inst)
            i += 1
        tcg_list = new_tcg_list

        addr = "????"
        call_inst = tcg_list[4]
        if call_inst.name == "call" and call_inst.args[0] == "uc_tracecode":
            # Unsure why this is, but in x86_64, I found that the first
            # instruction seems to be repeated, but with tmps instead of
            # the actual registers. It also seems like the second
            # argument to tracecode is 3 for the first call and 2 for
            # the others. I'll assume that 3's are always bad, and not
            # include those in the output
            if tcg_list[1].args[1] == "$0x3":
                return None

            addr_str = tcg_list[3].args[1]
            assert addr_str.startswith("$")
            addr = int(addr_str[1:], base=16)
            return TargetInst(addr, tcg_list[:5], tcg_list[5:])
        return TargetInst(addr, [], tcg_list)



# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================

from zelos import Zelos, HookType, IPlugin, CommandLineOption
from zelos.exceptions import InvalidRegException
from typing import List, Dict, Union
from collections import defaultdict

import functools
import os

from crashd.taint.tcg import TCGParse
from crashd.taint.taint_graph import TaintGraph
from crashd.dwarf.dwarf_source_code import (
    show_tainted_source,
    annotate_with_dwarf_data,
    get_function_info,
)

from zelos.ext.plugins.trace import Trace

_ = Trace

CommandLineOption(
    "taint",
    action="store_true",
    help=(
        "Enables collection of data that allows for taint tracking."
        " Collection will slow down overall run."
    ),
)

CommandLineOption(
    "taint_when",
    action="append",
    nargs="?",
    default=[],
    const="",
    metavar="ZML_STRING",
    help="Starts taint tracking when the specified ZML condition is met.",
)

CommandLineOption(
    "taint_output", default="", help="Specify taint output location.",
)

CommandLineOption(
    "source_code_path", default="", help="path to search for source code files"
)


class DataFlowException(Exception):
    pass


class TraceInfo:
    def __init__(self):
        self._trace: List[int] = []
        self._addr_comment: Dict[int, str] = {}
        """Map of idx to {memory_addr|reg: last_define_idx}"""
        self.uses = defaultdict(dict)
        """Map of idx -> {memory_addr|reg: value}"""
        self.defines = defaultdict(dict)

        """ Map of memory_addr|reg to last write idx"""
        self.last_write = {}
        self.label_overrides = {}

    @property
    def current_idx(self) -> int:
        return len(self._trace) - 1

    @property
    def last_addr(self) -> int:
        return self._trace[-1]

    def len(self) -> int:
        return len(self._trace)

    def add_address(self, addr: int):
        self._trace.append(addr)

    def add_use(
        self, addr_or_reg: Union[int, str], idx=None, label_override=None
    ):
        if idx is None:
            idx = self.current_idx
        self.uses[idx][addr_or_reg] = self.last_write.get(addr_or_reg, None)
        if label_override is not None:
            self.label_overrides[self.last_addr] = label_override

    def get_uses(self, idx: int) -> Dict[Union[int, str], int]:
        return self.uses[idx]

    def add_define(
        self,
        addr_or_reg: Union[int, str],
        value,
        idx=None,
        label_override=None,
    ):
        if idx is None:
            idx = self.current_idx
        if isinstance(value, int):
            value = hex(value)
        self.defines[idx][addr_or_reg] = value
        self.last_write[addr_or_reg] = idx
        if label_override is not None:
            self.label_overrides[self.last_addr] = label_override

    def get_defines(self, idx: int) -> Dict[Union[int, str], str]:
        return self.defines[idx]

    def _last_idx_of_address(self, address: int) -> int:
        # Find the last use of that address
        for idx in reversed(range(self.len())):
            if self._trace[idx] == address:
                return idx
        raise DataFlowException(f"Did not find target address {address:x}")


class DataFlow(IPlugin):
    """
    A collection of TargetInst. Dataflow techniques can be used on this
    collection to identify relationships between instructions

    TODO: Some instructions are conditional (movcond), meaning that even
    in this dynamic run, it may be unclear where each definition came
    from. We can deal with this multiple ways
      * Identify all possible sources it could have came from
      * Use the actual value that was given in order to give a guess on
        where the value came from.
    """

    def __init__(self, zelos: Zelos):
        super().__init__(zelos)
        self.dataflow_enabled = False
        if not zelos.config.taint and len(zelos.config.taint_when) == 0:
            return
        for zml_string in zelos.config.taint_when:
            zelos.internal_engine.zml_parser.trigger_on_zml(
                functools.partial(self.enable, zelos), zml_string,
            )
        if len(zelos.config.taint_when) == 0 and zelos.config.taint:
            self.enable(zelos)

    def enable(self, zelos: Zelos):
        """
        After calling this function, all dataflow in the target program
        will be tracked globally. The resulting flow is accessible
        through:
            - Dataflow.trace
            - Dataflow.ud_chain()
        """
        if self.dataflow_enabled:
            return
        self.dataflow_enabled = True
        self.trace = TraceInfo()
        self.define_use_map = {}
        self.reaching_defs: List[Dict[str, int]] = []

        def trace_inst(z, addr, size):

            self.trace.add_address(addr)
            # Sometimes the block hook doesn't work because the
            # definition of a block in tcg doesn't match up with the
            # size of the block. Need to look more into this.
            if addr not in self.define_use_map:
                self._update_tcg()
            # Delay adding register uses one address so that we can
            # get the value of the register after the instruction has
            # run.
            if self.trace.len() < 2:
                return
            last_addr = self.trace._trace[-2]
            (defs, _) = self.define_use_map[last_addr]
            idx = self.trace.len() - 2
            for register in defs:
                try:
                    reg_val = z.thread.get_reg(register)
                    self.trace.add_define(register, reg_val, idx=idx)
                except InvalidRegException:
                    # "uses" like "env" aren't registers, we may need
                    # To track them in the future though
                    pass

        def record_comments(z, addr, cmt):
            self.trace._addr_comment[addr] = cmt

        # This assumes memory hooks always run after instruction hooks.
        def trace_read(z, access, address, size, value):
            """ Add memory `use` to the trace """
            self.trace.add_use(address)

        def trace_write(z, access, address, size, value):
            """ Add memory `define` to the trace """
            self.trace.add_define(address, value)

        def trace_internal_read(z, access, address, size, value):
            """Add memory `use` to the trace originating from a zelos
            syscall emulation.
            """
            current_syscall = z.internal_engine.kernel._current_syscall
            if current_syscall is None:
                return
            for address in range(address, address + size):
                self.trace.add_use(address, label_override=current_syscall)

        def trace_internal_write(z, access, address, size, value):
            """Add memory `define` to the trace originating from a zelos
            syscall emulation.
            """
            current_syscall = z.internal_engine.kernel._current_syscall
            if current_syscall is None:
                return
            for i, address in enumerate(range(address, address + size)):
                self.trace.add_define(
                    address, value[i : i + 4], label_override=current_syscall
                )

        def trace_syscall(z, name, args, retval):
            """Add syscall arguments and return value uses and defines."""
            # FIXME: stack-based arguments will not be properly tracked
            #   here. For those, their addresses should be added. For
            #   example, syscalls with many arguments may put the
            #   overflow arguments on the stack. Uncommon.
            uses = set(z.internal_engine.kernel._REG_ARGS[: len(args._args)])
            defines = set([z.internal_engine.kernel._REG_RETURN])
            self.define_use_map[z.thread.getIP()] = (defines, uses)

        def trace_invalid_mem_access(z, access, address, size, value):
            """
            Handle invalid memory access violation by building a
            reverse taint graph from the crash site.
            """
            analyze_crash(self.zelos, self, self.trace.last_addr, address)
            return False

        zelos.plugins.trace.hook_comments(record_comments)
        zelos.hook_execution(HookType.EXEC.INST, trace_inst)
        zelos.hook_memory(HookType.MEMORY.READ, trace_read)
        zelos.hook_memory(HookType.MEMORY.WRITE, trace_write)
        zelos.hook_memory(HookType.MEMORY.INTERNAL_READ, trace_internal_read)
        zelos.hook_memory(HookType.MEMORY.INTERNAL_WRITE, trace_internal_write)
        zelos.hook_memory(HookType.MEMORY.INVALID, trace_invalid_mem_access)
        zelos.hook_syscalls(HookType.SYSCALL.AFTER, trace_syscall)

    def _update_tcg(self):
        """
        Attempts to get the tcg from the current address.
        """
        addr = self.zelos.thread.getIP()
        if addr in self.define_use_map:
            return
        insts = TCGParse().get_tcg(self.zelos)
        for i in insts:
            if self.state.arch == "x86":
                defines = self._adjust_x86_64_registers(i.defines())
                uses = self._adjust_x86_64_registers(i.uses())
            else:
                defines = i.defines()
                uses = i.uses()
            self.define_use_map[i.address] = (defines, uses)

    def _adjust_x86_64_registers(self, registers):
        return {
            "e" + reg[1:] if reg.startswith("r") else reg for reg in registers
        }

    def reverse_taint(self, start_addr: int):
        """ For testing """

        tg = TaintGraph(self, start_addr)
        return tg

    def _get_ida_func_name(self, address: int) -> str:
        if not hasattr(self, "names"):
            self.names = {
                n: a for a, n in self.zelos.plugins.ida.utils.Names()
            }

        return self.zelos.plugins.ida.idc.get_func_name(address)

    def _compute_defs(self):
        previous_dict = {}
        reaching_defs = [0] * len(self.trace._trace)
        for idx, addr in enumerate(self.trace._trace):
            (defs, _) = self.define_use_map.get(addr, ([], []))

            previous_dict.update({d: idx for d in defs})

            reaching_defs[idx] = previous_dict
            previous_dict = previous_dict.copy()
        self.reaching_defs = reaching_defs

    def ud_chain(self, target_idx: int) -> Dict[str, List[int]]:
        """
        Identifies the definition that reaches the current use.

        Returns:
            Dict of uses to the address of a potential definition.
        """
        if len(self.reaching_defs) == 0:
            self._compute_defs()
        ud_chain = defaultdict(list)
        target_address = self.trace._trace[target_idx]

        reaching_defs = self.reaching_defs[target_idx]

        (_, uses) = self.define_use_map.get(target_address, ([], []))
        uses = set(uses)  # we want a copy
        for u in uses:
            if u in ["env"]:
                continue
            reaching_def_idx = reaching_defs.get(u, None)
            if reaching_def_idx == target_idx:
                reaching_defs = self.reaching_defs[target_idx - 1]
                reaching_def_idx = reaching_defs.get(u, None)

            ud_chain[str(u)].append(reaching_def_idx)
        return ud_chain


def analyze_crash(z, dataflow, inst_address, mem_address):
    """
    Build a reverse taint graph from the crash site.
    """

    logger = dataflow.logger
    trace = dataflow.trace
    logger.notice("Execution finished.")
    taint_path = TaintGraph(dataflow, inst_address, mem_address)
    zelos_module_base = z.internal_engine.memory.get_module_base(
        z.target_binary_path
    )
    logger.info("Parsing DWARF info")
    annotate_with_dwarf_data(z, z.target_binary_path, trace._trace, taint_path)
    taint_path._addr2func = get_function_info(
        z.target_binary_path, taint_path, zelos_module_base
    )
    try:
        logger.info("Creating source taint graph")
        show_tainted_source(z, z.target_binary_path, trace._trace, taint_path)
        zcovPath = os.path.abspath(
            os.path.join(z.config.source_code_path, "crashd.zcov")
        )
        logger.notice(f"Wrote file: {zcovPath}")
    except Exception:
        logger.exception("Unable to show source code")
    if True:
        # This is all png graph generation
        from .render.graphviz import (
            render_reduced_path_taint_graph,
            render_path_taint_graph,
        )

        render_reduced_path_taint_graph(logger, taint_path)
        render_path_taint_graph(logger, taint_path, trace)

        # self.get_ida_taint_overlay(
        #     z.internal_engine.main_module.EntryPoint,
        #     trace,
        #     taint_path,
        #     open(crash_trace_file_name, "w"),
        # )
    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import codecs
import os
import re
import sys

from shutil import rmtree

from setuptools import Command, find_packages, setup

NAME = "zelos-crashd"
META_PATH = os.path.join("src", "crashd", "__init__.py")
PACKAGES = find_packages(where="src")
KEYWORDS = [
    "emulation",
    "dynamic analysis",
    "binary analysis",
    "zelos",
    "crash",
    "vulnerability assessment",
]
PROJECT_URLS = {
    "Bug Tracker": "https://github.com/zeropointdynamics/zelos-crashd/issues",
    "Source Code": "https://github.com/zeropointdynamics/zelos-crashd",
}
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Natural Language :: English",
    "License :: OSI Approved :: GNU Affero General Public License v3",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
]
INSTALL_REQUIRES = [
    "zelos>=0.2.0",
    "pyelftools@git+git://github.com/eliben/pyelftools.git@7ca16680f4eb01e6b6bbbc222f18de56460426d1",
]

########################################################################

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of
    the resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


VERSION = find_meta("version")
URL = find_meta("url")
LONG = (
    read("README.md")
    + "\n\n"
    + read("CHANGELOG.md")
    + "\n\n"
    + read("AUTHORS.md")
)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(HERE, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system(
            "{0} setup.py sdist bdist_wheel --universal".format(sys.executable)
        )

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(VERSION))
        os.system("git push --tags")

        sys.exit()


if __name__ == "__main__":
    setup(
        name=NAME,
        description=find_meta("description"),
        license=find_meta("license"),
        url=URL,
        project_urls=PROJECT_URLS,
        version=VERSION,
        author=find_meta("author"),
        author_email=find_meta("email"),
        maintainer=find_meta("author"),
        maintainer_email=find_meta("email"),
        keywords=KEYWORDS,
        long_description=LONG,
        long_description_content_type="text/markdown",
        packages=PACKAGES,
        package_dir={"": "src"},
        python_requires=">=3.6.0",
        zip_safe=False,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        include_package_data=True,
        options={"bdist_wheel": {"universal": "1"}},
        setup_requires=["wheel"],
        cmdclass={"upload": UploadCommand},
        entry_points={
            "zelos.plugins": [
                "asan=crashd.asan",
                "dataflow=crashd.taint",
                "ida=crashd.static_analysis",
            ],
        },
    )



# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================

# Python2 Script for Ida

from idautils import *
from idaapi import *
import idc
import pickle
import base64
import types


def write_data(out, data):
    data = base64.b64encode(pickle.dumps(data)) + "\n"
    out.write(data)
    out.flush()


in_path = idc.ARGV[1]
out_path = idc.ARGV[2]

p = open(in_path, "r")
out = open(out_path, "wb")

Wait()
ea = BeginEA()
result = None
while True:
    try:
        cmd = p.readline()
        exec_command = "global result; result = %s" % cmd
        try:
            exec_command = "global result; result = %s" % cmd
            exec(exec_command)
        except Exception as e:
            write_data(out, str(e) + " | exec_command : " + exec_command)
        else:
            if isinstance(result, types.GeneratorType):
                result = list(result)
            write_data(out, result)
    except Exception as e:
        write_data(out, "Error in plugin: " + str(e))

idc.Exit(0)



# Copyright (C) 2020 Zeropoint Dynamics

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
# ======================================================================

from collections import defaultdict
from termcolor import colored
import os
import html
from pathlib import Path
from typing import Optional

from elftools.dwarf.descriptions import describe_form_class
from elftools.elf.elffile import ELFFile


def find_file(path, source_code_path):
    """
    First checks the current directory for the path.
    Then checks recursively in the source code folder for the path
    """
    if os.path.exists(path):
        return path
    matching_files = [path for path in Path(source_code_path).rglob(path)]
    if len(matching_files) == 0:
        # print(
        #     f"Could not find source code file {path} within {source_code_path}"
        # )
        return None
    if len(matching_files) > 1:
        # print(
        #     f"There is more than one matching file for {path}: {matching_files}. Picking {matching_files[0]}"
        # )
        pass
    return matching_files[0]


def _construct_zcov_files(
    taint_path, address_map, trace, edges, files, file_to_syspath, src_path
):
    from pathlib import Path
    import json

    crash_line = None
    crash_file = None
    zcov_files = defaultdict(dict)

    # Merge values used for each source code line of dataflow
    for k, v in taint_path.reduced_path.items():
        k_addr = k
        file, line_num = address_map.get(k_addr, (None, None))

        if file not in files:
            continue

        # Get absolute paths, resolve any symbolic links, etc.
        parent_path = os.path.realpath(src_path)
        child_path = os.path.realpath(file_to_syspath[file])

        # Ensure file path is actually rooted in the src parent path
        # (i.e. ignore system libc headers, etc.)
        if os.path.commonpath([parent_path]) != os.path.commonpath(
            [parent_path, child_path]
        ):
            continue

        line_num += 1
        assembly = taint_path._assembly.get(k_addr, "Missing Assembly")
        if crash_line is None:
            kind = "FLOW_END"
            crash_line = line_num
            crash_file = file
            # if taint_path._dataflow.zelos.config.asan:
            #     crash_summary = (
            #         taint_path._dataflow.zelos.plugins.asan.get_crash_summary()
            #     )
            #     crash_summary_comments = "".join(
            #         [f";{line}\n" for line in crash_summary.split("\n")]
            #     )
            #     assembly = crash_summary_comments + assembly
        else:
            kind = "FLOW_THROUGH"
        vals_used = list(set([str(rtd[1].val) for rtd in v.items()]))
        # print(file, line_num, kind, vals_used)
        if int(line_num) in zcov_files[file]:
            existing_vals = zcov_files[file][int(line_num)]["meta"]
            zcov_files[file][int(line_num)]["meta"] = list(
                set(existing_vals + vals_used)
            )
        else:
            zcov_files[file][int(line_num)] = {
                "kind": kind,
                "meta": vals_used,
                "asm": [],
            }
        zcov_files[file][int(line_num)]["asm"].append((k_addr, assembly))

    # print("  zcov merge values")
    # Add lines-executed information
    count = 0
    after_dataflow = True
    addrs = set()
    for addr in reversed(trace):
        file, line_num = address_map.get(addr, (None, None))
        if file not in zcov_files:
            continue
        # Get absolute paths, resolve any symbolic links, etc.
        parent_path = os.path.realpath(src_path)
        child_path = os.path.realpath(file)

        # Ensure file path is actually rooted in the src parent path
        # (i.e. ignore system libc headers, etc.)
        if os.path.commonpath([parent_path]) != os.path.commonpath(
            [parent_path, child_path]
        ):
            continue
        line_num += 1
        if line_num == crash_line:
            after_dataflow = False
        if int(line_num) not in zcov_files[file]:
            if after_dataflow:
                zcov_files[file][int(line_num)] = {
                    "kind": "EXEC_AFTER_FLOW_END",
                    "asm": [],
                }
            else:
                zcov_files[file][int(line_num)] = {"kind": "EXEC", "asm": []}
        if addr in addrs:
            continue
        addrs.add(addr)
        assembly = taint_path._assembly.get(addr, None)
        if assembly is None:
            assembly = taint_path.get_assembly_for_range(addr, addr + 20)[addr]
        zcov_files[file][int(line_num)]["asm"].append((addr, assembly))
        count += 1

    # Add annotation of where asan guarded memory was triggered.
    # if taint_path._dataflow.zelos.plugins.asan._asan_guard_triggered:
    #     addr = (
    #         taint_path._dataflow.zelos.plugins.asan.crash_alloc_info.inst_address
    #     )
    #     file, line_num = address_map.get(addr, (None, None))
    #     line_num += 1
    #     if file in zcov_files:
    #         zcov_files[file][line_num]["kind"] = "ALLOC"

    # print("  zcov added lines executed")
    # Add data flow line edge information
    for src, dests in edges.items():
        srcfile, srcline_num = address_map.get(src, (None, None))
        if srcfile not in zcov_files:
            continue
        srcline_num += 1
        for dest in dests:
            destfile, destline_num = address_map.get(dest, (None, None))
            if destfile not in zcov_files:
                continue
            destline_num += 1

            # print(f"  {destfile}{destline_num} -> {srcfile}{srcline_num}")
            if destfile not in zcov_files or srcfile not in zcov_files:
                continue
            if "data_from" not in zcov_files[destfile][int(destline_num)]:
                zcov_files[destfile][int(destline_num)]["data_from"] = list()
            if "data_to" not in zcov_files[srcfile][int(srcline_num)]:
                zcov_files[srcfile][int(srcline_num)]["data_to"] = list()
            zcov_files[destfile][int(destline_num)]["data_from"].append(
                {"file": srcfile, "line_number": srcline_num}
            )
            zcov_files[srcfile][int(srcline_num)]["data_to"].append(
                {"file": destfile, "line_number": destline_num}
            )
            # print(f"{destfile}{destline_num} -> {srcfile}{srcline_num}")
    # print("  zcov add data flow edges")
    # Generate zcov-formatted JSON
    single_json = {"files": list(), "graphs": list()}
    for file, zcov_content in zcov_files.items():
        if file not in file_to_syspath:
            continue
        zcov_file = Path(file_to_syspath[file]).with_suffix(".zcov")
        zcov_json = defaultdict(dict)
        zcov_json["file"] = file
        lines = []
        for line_num, line_info in zcov_content.items():
            line = {}
            line["line_number"] = line_num
            line["kind"] = line_info["kind"]
            if "data_to" in line_info:
                line["data_from"] = line_info["data_to"]
            if "data_from" in line_info:
                line["data_to"] = line_info["data_from"]
            if "meta" in line_info:
                # Convert vals list to short string
                vals_used = line_info["meta"]
                vals_used = sorted(
                    vals_used, key=lambda x: 0 if "=" in x else 1
                )
                elipsis = ""
                if len(vals_used) > 5:
                    elipsis = f",(+{len(vals_used)-5} more)"
                    vals_used = vals_used[:5]
                info_str = "Vals: " + ",".join(vals_used) + elipsis
                line["meta"] = info_str
            if "asm" in line_info:
                # sorted_lines = sorted(line_info["asm"], key=lambda x: x[0])
                # line["asm"] = [x[1] for x in sorted_lines]
                line["asm"] = [x[1] for x in reversed(line_info["asm"])]
            lines.append(line)
        zcov_json["lines"] = lines
        # json_str = json.dumps(zcov_json, indent=2, sort_keys=True)
        single_json["files"].append(zcov_json)
        # print(f"==== {zcov_file} ====")
        # print(json_str)
        # with open(zcov_file, "w") as f:
        #     f.write(json_str)

    (
        src_graph,
        ordering,
        wave_order,
        child2parents,
        parent2children,
    ) = export_source_graph(taint_path, address_map, files)

    num_waves = max(wave_order.values()) + 1
    wave_ordering = [[] for _ in range(num_waves)]
    for key, wave_num in wave_order.items():
        wave_ordering[wave_num].append(key)

    graph_json = {
        "name": "source_graph",
        "data": src_graph,
        "crashpoint": f"{crash_file}{str(crash_line - 1)}",
        "ordering": ordering,
        "wave_ordering": wave_ordering,
        "child2parents": child2parents,
        "parent2children": parent2children,
    }
    single_json["graphs"].append(graph_json)

    with open(os.path.join(src_path, "crashd.zcov"), "w") as f:
        f.write(json.dumps(single_json, indent=2, sort_keys=True))
    _cwd = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(_cwd, "template.html"), "r") as inf:
        with open(os.path.join(src_path, "crashd.graph.html"), "w") as outf:
            data = inf.read()
            data = data.replace(
                "@@ZPD_GRAPH@@", json.dumps(single_json["graphs"])
            )
            outf.write(data)


def annotate_with_dwarf_data(zelos, binary_path, trace, taint_path):
    zelos_module_base = zelos.internal_engine.memory.get_module_base(
        binary_path
    )
    address_map = process_file(binary_path, zelos_module_base)
    dwarf_data = DwarfData(address_map, zelos.config.source_code_path)
    # TODO: The following uses the updated DwarfData
    # dwarf_data = DwarfData(
    #     binary_path, zelos.config.source_code_path, zelos_module_base
    # )
    taint_path._dwarf_data = dwarf_data


def get_nodes_and_edges(taint_graph, address_map):
    # print("Setting up source graph")
    # tuples containing source code addr and next addr to check
    open_paths = [(taint_graph._start_addr, taint_graph._start_addr)]
    # No need to analyze an address that has already been analyzed.
    analyzed_addrs = set()
    # use to defines
    edges = defaultdict(list)
    nodes = set()
    while len(open_paths) > 0:
        child, next_ancestor = open_paths.pop()
        if next_ancestor in analyzed_addrs:
            continue
        analyzed_addrs.add(next_ancestor)
        parents = taint_graph._reduced_path_parents.get(next_ancestor, [])
        for parent in parents:
            if address_map.get(parent, None) is not None:
                edges[child].append(parent)
                nodes.add(child)
                nodes.add(parent)
                if parent not in analyzed_addrs:
                    open_paths.append((parent, parent))
            else:
                if parent not in analyzed_addrs:
                    open_paths.append((child, parent))
    return (nodes, edges)


def show_tainted_source(zelos, binary_path, trace, taint_path):
    dwarf_data = taint_path._dwarf_data
    files = dwarf_data._file_lines
    file_to_syspath = dwarf_data._file_to_syspath
    address_map = dwarf_data._address_map
    address_map._attach_src_to_external_addrs(trace, dwarf_data)

    from crashd.taint.render.graphviz import render_source_graph

    render_source_graph(taint_path, address_map, files)

    (nodes, edges) = get_nodes_and_edges(taint_path, address_map)
    if len(file_to_syspath) > 0:
        _construct_zcov_files(
            taint_path,
            address_map,
            trace,
            edges,
            files,
            file_to_syspath,
            zelos.config.source_code_path,
        )

    if not zelos.config.taint_output == "terminal":
        return

    # print("generated .zcov file")
    crash_line = None
    source_path = {}
    changed_file_lines = defaultdict(list)
    for k, v in taint_path.reduced_path.items():
        k_addr = k
        file, line = address_map.get(k_addr, (None, None))
        source_path[(file, line)] = v
        if file in files:
            if crash_line is None:
                files[file][line] = colored(
                    # f"!0x{k:x}" +
                    files[file][line] + str(v) + "\n",
                    color="red",
                )
                crash_line = line
                changed_file_lines[file].append(line)
            else:
                files[file][line] = colored(
                    # f"*0x{k:x}" +
                    files[file][line] + str(v) + "\n",
                    color="white",
                    attrs=["bold"],
                )
                changed_file_lines[file].append(line)

    # print("Updated source lines with data flow")
    count = 0
    # Get first trace line that is in the source
    for addr in reversed(trace):
        file, line = address_map.get(addr, (None, None))
        if file not in files:
            continue
        if line == crash_line:
            break
        files[file][line] = colored(
            files[file][line] + f" Count: {count}\n",
            color="green",
            attrs=["bold"],
        )
        changed_file_lines[file].append(line)
        count += 1

    # print("Updated source lines with trace to crash")

    if len(changed_file_lines) == 0:
        print(
            "There are no lines in source that correspond to the taint path."
            " There may have been a bug."
        )

    for path, lines in files.items():
        if path not in changed_file_lines:
            continue
        idxes = changed_file_lines[path]
        # print(path, idxes)
        lines_to_print = []
        indices_chosen = set()
        context = 3
        for idx in reversed(idxes):
            i = idx - context
            while i < idx + context + 1:
                if i < 0 or i >= len(lines):
                    i += 1
                    continue
                if i in indices_chosen:
                    i += 1
                    continue
                if i > idx and i in idxes:
                    idx = i
                lines_to_print.append(lines[i])
                indices_chosen.add(i)
                i += 1
            if lines_to_print[-1] != "...\n":
                lines_to_print.append("...\n")
        print("".join(lines_to_print))

    # print("Changed lines: ")
    # print(changed_file_lines)
    # print("Num lines: ", len(changed_file_lines))


def process_file(filename, zelos_module_base):
    """
    Returns a dictionary containing addr -> (file, 0-indexed source code line)
    Returns 0 indexed results, while ElfFile gives 1 indexed results
    """
    # print("Processing file:", filename)
    with open(filename, "rb") as f:
        elffile = ELFFile(f)
        if not elffile.has_dwarf_info():
            # print("  file has no DWARF info")
            return

        symbols_module_base = _get_module_base(elffile)
        offset = zelos_module_base - symbols_module_base
        if offset != 0x1000:  # TODO TEMP for static binaries
            offset = 0
        # print("Got offset: ", offset)
        total_address_map = {}
        dwarfinfo = elffile.get_dwarf_info()

        return decode_file_lines(dwarfinfo, offset)


def __extract_value(attr, key, default):
    if key in attr.keys():
        return attr[key].value
    return default


def __handle_DIE(DIE):
    # we are interested in two things: name and address range
    tag = DIE.tag
    attr = DIE.attributes
    # ignore compile unit info
    if tag == "DW_TAG_compile_unit":
        return None
    # check for low_pc
    lowpc = __extract_value(attr, "DW_AT_low_pc", -1)
    # we don't care if DIE holds no address
    if lowpc == -1:
        return None
    elif "DW_AT_high_pc" in attr.keys():
        highpc_attr = attr["DW_AT_high_pc"]
        highpc_attr_class = describe_form_class(highpc_attr.form)
        if highpc_attr_class == "address":
            highpc = highpc_attr.value
        elif highpc_attr_class == "constant":
            highpc = lowpc + highpc_attr.value
    else:
        highpc = lowpc
    # recursive search for name
    current_die = DIE
    while True:
        name = __extract_value(attr, "DW_AT_name", b"")
        if name and current_die.tag == "DW_TAG_subprogram":
            return {"name": name.decode(), "range": (lowpc, highpc)}
        origin = __extract_value(attr, "DW_AT_abstract_origin", -1)
        if origin == -1:
            break
        current_die = current_die.get_DIE_from_attribute(
            "DW_AT_abstract_origin"
        )
        attr = current_die.attributes
    return None


def get_function_info(filename, taint_graph, zelos_module_base):
    """
    Returns mapping of function names to tuple (low addr, high addr)
    """
    # print("Parsing DWARF info")
    with open(filename, "rb") as f:
        elffile = ELFFile(f)
        if not elffile.has_dwarf_info():
            # print("  file has no DWARF info")
            return

        symbols_module_base = _get_module_base(elffile)
        offset = zelos_module_base - symbols_module_base
        if offset != 0x1000:  # TODO TEMP for static binaries
            offset = 0
        # print("Got offset: ", offset)
        dwarfinfo = elffile.get_dwarf_info()
        functions = defaultdict(list)
        entries = []
        for CU in dwarfinfo.iter_CUs():
            for DIE in CU.iter_DIEs():
                die_info = __handle_DIE(DIE)
                if die_info is not None:
                    entries.append(die_info)
        for entry in entries:
            func_name = entry["name"]
            low, high = entry["range"]
            functions[func_name].append((low + offset, high + offset))
    # print("  Constructing addr2func")
    addr2func = {}
    for addr in taint_graph._reduced_path.keys():
        for f, ranges in functions.items():
            for r in ranges:
                (low, high) = r
                if low <= addr <= high:
                    addr2func[addr] = f
    # print("Done parsing dwarf info")
    return addr2func


class AddressEntry:
    def __init__(self, low, high, file, line_num):
        self.low = low
        self.high = high
        self.file = file
        self.line_num = line_num

    def __str__(self):
        return f"(0x{self.low:x}-0x{self.high:x}: {self.file}:{self.line_num})"


class AddressMap:
    def __init__(self):
        self._address_low_high = []
        self._cache = {}
        self._source2addr_range = {}

    def __str__(self):
        return [entry.__str__() for entry in self._address_low_high].__str__()

    def files(self):
        return {entry.file for entry in self._address_low_high}

    def get_addr_range_from_source(self, file, line_num):
        return self._source2addr_range.get(f"{file}{line_num}", (None, None))

    def _add_cache(self, addr, file, line):
        self._cache[addr] = (file, line - 1)

    def add(self, low, high, file, line):
        file = file.decode()
        self._add_cache(low, file, line)
        self._source2addr_range[f"{file}{line}"] = (low, high)
        for addr in range(low, high):
            self._add_cache(addr, file, line)

        entry = AddressEntry(low, high, file, line)
        self._address_low_high.append(entry)
        self._address_low_high.sort(key=lambda x: x.low)

    def get(self, addr: int, default=(None, None)):
        cached_val = self._cache.get(addr, None)
        if cached_val is not None:
            return cached_val

        idx = self.binary_search(addr)
        if idx is not None:
            entry = self._address_low_high[idx]
            retval = (entry.file, entry.line_num - 1)
            self._cache[addr] = retval
            return retval

        return default

    def _attach_src_to_external_addrs(self, trace, dwarf_data):
        """
        Certain addresses in the trace are not associated with source
        code because they are in external modules. We associate those
        addresses with source code from the last line that occurred
        before moving to the external module.
        """
        # External modules can call each other. This means that the same
        # address could have multiple candidates for what address should
        # be associated with them.
        files = dwarf_data._file_lines

        current_source = None
        current_file = None
        current_line = None
        smeared = {}
        for addr in trace:
            file, line_num = self.get(addr)

            if file in files and addr not in smeared:
                dwarf_data._addr2source[addr] = files[file][line_num]
                current_source = files[file][line_num]
                current_file = file
                current_line = line_num
            elif current_source is not None:
                dwarf_data._addr2source[addr] = "within " + current_source
                self._add_cache(addr, current_file, current_line + 1)
                smeared[addr] = current_source

    def binary_search(self, x):
        low = 0
        mid = 0
        high = len(self._address_low_high) - 1
        while low <= high:
            mid = (high + low) // 2
            entry = self._address_low_high[mid]
            if entry.low <= x < entry.high:
                return mid
            if entry.low == entry.high and entry.low == x:
                return mid

            # Check if x is present at mid
            if entry.low < x:
                low = mid + 1

            # If x is greater, ignore left half
            elif entry.low > x:
                high = mid - 1

        # If we reach here, then the element was not present
        return None


class DwarfData:
    def __init__(self, address_map, source_code_path):
        self._address_map = address_map
        self._file_lines = {}
        self._file_to_syspath = {}
        self._addr2source = {}

        self.setup_file_lines(source_code_path)

    def setup_file_lines(self, source_code_path):
        for filename in self._address_map.files():
            located_filename = find_file(filename, source_code_path)
            if located_filename is None:
                continue
            f = open(located_filename, "r")
            self._file_to_syspath[filename] = located_filename
            # Keep this keyed by the original file name so that later on we
            # can find these lines
            self._file_lines[filename] = list(f.readlines())

    def get_source(self, addr: int) -> Optional[str]:
        return self._addr2source.get(addr, None)


def decode_file_lines(dwarfinfo, offset):
    address_map = AddressMap()
    # Go over all the line programs in the DWARF information, looking for
    # one that describes the given address.
    for CU in dwarfinfo.iter_CUs():
        # First, look at line programs to find the file/line for the address
        lineprog = dwarfinfo.line_program_for_CU(CU)
        prevstate = None
        for entry in lineprog.get_entries():
            # We're interested in those entries where a new state is assigned
            if entry.state is None:
                continue
            if entry.state.end_sequence:
                # if the line number sequence ends, clear prevstate.
                prevstate = None
                continue
            # Looking for a range of addresses in two consecutive states that
            # contain the required address.
            if prevstate:
                filename = lineprog["file_entry"][prevstate.file - 1].name
                line = prevstate.line
                address_map.add(
                    prevstate.address + offset,
                    entry.state.address + offset,
                    filename,
                    line,
                )
            prevstate = entry.state
    return address_map


def _get_module_base(elf_file):
    # This function DOES NOT WORK for static binaries.
    # It gets lucky on binaries where the main module is loaded at 0,
    # which is what this function normally returns. Have to find a way
    # to get the desired load address of binaries. Maybe one way to get
    # around is always return 0 for dynamic binaries.
    segment_addrs = [s.header.p_vaddr for s in elf_file.iter_segments()]
    # print(f"Segment_addrs: {segment_addrs}")
    return min(segment_addrs)


# ELK Graph Functions


def _calc_width(text):
    return len(text) * 8 + 20


def _create_node(_file, _line, _text):
    _id = f"{_file}{_line}"
    node = {
        "id": _id,
        "labels": [{"id": f"{_id}_label", "text": _text, "x": 10, "y": 4,}],
        "width": _calc_width(_text),
        "height": 24,
        "file": f"{_file}",
        "line": _line,
    }
    return node


def _create_group(_id, label=None):
    group = {
        "id": f"group_{_id}",
        "children": [],
        "edges": [],
        "layoutOptions": {"elk.direction": "DOWN"},
    }
    if label:
        group["labels"] = [
            {
                "id": f"group_{_id}_label",
                "text": f"{html.escape(label)}",
                "width": _calc_width(label),
                "height": 24,
            }
        ]
    return group


def _create_edge(srcfile, srcline, destfile, destline):
    _src = f"{srcfile}{srcline}"
    _dest = f"{destfile}{destline}"
    _id = f"edge_{_src}{_dest}"
    edge = {
        "id": _id,
        "source": _dest,
        "target": _src,
        "sourceFile": destfile,
        "targetFile": srcfile,
        "sourceLine": destline,
        "targetLine": srcline,
    }
    return edge


def _find_group(edge, groups):
    # Given an edge, search for a function-group
    # that contains both it's source and target
    # nodes.
    src_group = None
    dest_group = None
    for fn, g in groups.items():
        for c in g.get("children", []):
            if edge["sourceFile"] == c.get("file", None) and edge[
                "sourceLine"
            ] == c.get("line", None):
                src_group = fn
            if edge["targetFile"] == c.get("file", None) and edge[
                "targetLine"
            ] == c.get("line", None):
                dest_group = fn
            if src_group is not None and dest_group is not None:
                if src_group == dest_group:
                    return src_group
                return None
    return None


def _find_fn(file, line, groups):
    # Given a file and line number, search for
    # it's containing function-group
    for fn, g in groups.items():
        for c in g.get("children", []):
            if file == c.get("file", None) and line == c.get("line", None):
                return fn
    return None


def _find_edges(node, groups):
    # Given a node that is not part of a group,
    # search for a function-group that contains
    # edges to or from it.
    for fn, g in groups.items():
        for e in g.get("edges", []):
            if (
                node["file"] == e["sourceFile"]
                or node["file"] == e["targetFile"]
            ) and (
                node["line"] == e["sourceLine"]
                or node["line"] == e["targetLine"]
            ):
                return fn
    return None


def export_source_graph(taint_path, address_map, files):
    # Helper function for getting id for the source graph
    def get_node_id(address_map, addr):
        (file, line_num) = address_map.get(addr, (None, None))
        if file is None:
            return None
        return f"{file}{line_num}"

    graph_boy = {
        "id": "root",
        "layoutOptions": {
            "algorithm": "layered",
            "elk.direction": "DOWN",
            "hierarchyHandling": "INCLUDE_CHILDREN",
        },
        "children": [],
        "edges": [],
    }
    # tuples containing source code addr and next addr to check
    open_paths = [(taint_path._start_addr, taint_path._start_addr)]
    # No need to analyze an address that has already been analyzed.
    analyzed_addrs = set()
    # use to defines
    edges = defaultdict(list)
    nodes = set()
    # The order of nodes
    ordering = []
    # Wave order is broken if you do analysis in depth first search.
    # The algorithm for calculating wave order only works if you use
    # breadth first.
    wave_order = {get_node_id(address_map, taint_path._start_addr): 0}
    child2parents = defaultdict(list)
    parent2children = defaultdict(list)
    while len(open_paths) > 0:
        child, next_ancestor = open_paths.pop()
        if next_ancestor in analyzed_addrs:
            continue
        if child not in taint_path._addr2func:
            if taint_path._dataflow.zelos.config.link_ida is not None:
                taint_path._addr2func[
                    child
                ] = taint_path._dataflow._get_ida_func_name(child)

        analyzed_addrs.add(next_ancestor)

        ancestor_id = get_node_id(address_map, next_ancestor)
        if ancestor_id not in ordering:
            ordering.append(ancestor_id)

        parents = taint_path._reduced_path_parents.get(next_ancestor, [])
        for parent in parents:
            (file, line_num) = address_map.get(parent, (None, None))
            if file is None:
                if parent not in analyzed_addrs:
                    open_paths.append((child, parent))
                continue

            edges[child].append(parent)
            nodes.add(child)
            nodes.add(parent)
            child_id = get_node_id(address_map, child)
            parent_id = get_node_id(address_map, parent)
            if parent_id not in wave_order:
                wave_order[parent_id] = wave_order[child_id] + 1
            if parent not in analyzed_addrs:
                open_paths.append((parent, parent))

    groups = dict()
    added_nodes = set()
    function_map = taint_path._addr2func
    line2func = defaultdict(lambda: defaultdict(None))
    for n in nodes:
        file, line_num = address_map.get(n, (None, None))
        if file not in files:
            continue
        function = function_map.get(n, None)
        if function is not None:
            if function and function not in groups.keys():
                fn_label = f"{file}::{function}"
                groups[function] = _create_group(function, label=fn_label)
            if file and line_num:
                line2func[file][line_num] = function

        src_line = files[file][line_num].strip()
        fline = f"{file}{line_num}"
        if fline in added_nodes:
            continue
        added_nodes.add(fline)
        # add node
        node = _create_node(file, line_num, " ".join(src_line.split()))
        if function:
            groups[function]["children"].append(node)
        else:
            graph_boy["children"].append(node)

    added_edges = set()
    for src, dests in edges.items():
        srcfile, srcline_num = address_map.get(src, (None, None))
        srcfn = function_map.get(src, None)
        if srcfn is None:
            srcfn = line2func[srcfile].get(srcline_num, None)
        for dest in dests:
            destfile, destline_num = address_map.get(dest, (None, None))
            destfn = function_map.get(dest, None)
            if destfn is None:
                destfn = line2func[destfile].get(destline_num, None)
            src = f"{srcfile}{srcline_num}"
            dest = f"{destfile}{destline_num}"
            if src == "NoneNone" or dest == "NoneNone":
                continue
            edge_name = f"edge_{src}{dest}"
            if edge_name in added_edges:
                continue
            if src not in added_nodes or dest not in added_nodes:
                continue

            child2parents[src].append((edge_name, dest))
            parent2children[dest].append((edge_name, src))

            added_edges.add(edge_name)
            edge = _create_edge(srcfile, srcline_num, destfile, destline_num)
            fn = _find_group(edge, groups)
            if fn is not None:
                groups[fn]["edges"].append(edge)
            # Same function
            elif srcfn and srcfn == destfn:
                groups[srcfn]["edges"].append(edge)
            # Different function
            elif srcfn and destfn and srcfn != destfn:
                graph_boy["edges"].append(edge)
            # could be either
            else:
                # if a sourcefile is known and the same
                if (srcfile or destfile) and srcfile == destfile:
                    # if the line number same
                    if (
                        srcline_num or destline_num
                    ) and srcline_num == destline_num:
                        # If line num same and one is mapped to a fn
                        # it's safe to assume same fn
                        if srcfn is not None and destfn is None:
                            groups[srcfn]["edges"].append(edge)
                        elif srcfn is None and destfn is not None:
                            groups[destfn]["edges"].append(edge)
                        else:
                            # If line num is same but neither mapped to a fn
                            # Check if we've seen this address mapped to a fn
                            # previously.
                            fn = _find_fn(srcfile, srcline_num, groups)
                            if fn is not None:
                                groups[fn]["edges"].append(edge)
                            else:
                                graph_boy["edges"].append(edge)
                    else:
                        # if different line number, check that we've seen both previously
                        sfn = _find_fn(srcfile, srcline_num, groups)
                        dfn = _find_fn(destfile, destline_num, groups)
                        if (sfn or dfn) and sfn == dfn:
                            groups[sfn]["edges"].append(edge)
                        else:
                            graph_boy["edges"].append(edge)
                else:
                    graph_boy["edges"].append(edge)

    # Take care of any incorrect func mappings
    tmp = []
    for c in graph_boy["children"]:
        if c.get("file", None) and c.get("line", None):
            fn = _find_edges(c, groups)
            if fn is not None:
                groups[fn]["children"].append(c)
            else:
                tmp.append(c)
    graph_boy["children"] = tmp

    # Take care of any incorrect edges
    tmp = []
    for e in graph_boy["edges"]:
        fn = _find_group(e, groups)
        if fn is not None:
            groups[fn]["edges"].append(e)
        else:
            tmp.append(e)
    graph_boy["edges"] = tmp

    for fn in groups.keys():
        graph_boy["children"].append(groups[fn])

    # import json
    # with open("output.json", "w") as fp:
    #     json.dump(graph_boy, fp)
    return graph_boy, ordering, wave_order, child2parents, parent2children
